export interface FormatOptions {
	/** Whether or not to display the fellow's email */
	displayEmail?: true
	/** Whether or not to display the copright icon */
	displayCopyright?: boolean
	/** Whether or not to display the copyright years */
	displayYears?: boolean
	/** Whether or not to display a link to the user's contributions, if used along with {@link .githubRepoSlug} */
	displayContributions?: boolean
	/** The repository for when using with {@link .displayContributions} */
	githubRepoSlug?: string
	/** An array of fields to prefer for the URL */
	urlFields?: ['githubUrl', 'url']
}
/** Comparator for sorting fellows in an array */
export declare function comparator(a: Fellow, b: Fellow): 0 | 1 | -1
/** A fellow with similarties to other people */
export default class Fellow {
	/** Note that any property can be assigned directly to fellow */
	[key: string]: any
	/** Actual name that is stored, otherwise falls back to username from the url fields */
	private _name?
	/** Years active for the current repository, extracted from the name */
	years?: string
	/** URLs used */
	readonly urls: Set<string>
	/** Emails used */
	readonly emails: Set<string>
	/** Map of repository slugs with the contributions from the user */
	readonly contributions: Map<string, number>
	/** Set of repository slugs that the fellow administers to */
	readonly administeredRepositories: Set<string>
	/** Set of repository slugs that the fellow contributes to */
	readonly contributedRepositories: Set<string>
	/** Set of repository slugs that the fellow maintains */
	readonly maintainedRepositories: Set<string>
	/** Set of repository slugs that the fellow authors */
	readonly authoredRepositories: Set<string>
	/**
	 * An array of field names that are used to determine if two fellow's are the same.
	 * Don't need to add usernames for github, twitter, and facebook, as they will be compared via `urls`.
	 * Can't compare just username, as that is not unique unless comparison's are on the same service, hence why `urls` are used and not `usernames`.
	 */
	protected readonly idFields: string[]
	/** An array of field names that are used to determine the fellow's URL */
	protected readonly urlFields: string[]
	/** A singleton attached to the class that stores it's instances to enable convergence of data */
	static readonly fellows: Array<Fellow>
	/**
	 * Sort a list of fellows.
	 * Uses {@link Fellow::sort} for the comparison.
	 */
	static sort(list: Array<Fellow>): Fellow[]
	/** Flatten lists of fellows into one set of fellows */
	static flatten(lists: Array<Array<Fellow> | Set<Fellow>>): Set<Fellow>
	/** Compare to another fellow for sorting. */
	compare(other: Fellow): -1 | 0 | 1
	/**
	 * Compare to another fellow for equivalancy.
	 * Uses {@link Fellow::idFields} for the comparison.
	 * @param other The other fellow to compare ourselves with
	 * @returns Returns `true` if they appear to be the same person, or `false` if not.
	 */
	same(other: Fellow): boolean
	/**
	 * With the value, see if an existing fellow exists in our singleton list property with the value, otherwise create a new fellow instance with the value and add them to our singleton list.
	 * Uses {@link Fellow::same} for the comparison.
	 * @param value The value to create a new fellow instance or find the existing fellow instance with
	 * @param add Whether to add the created person to the list
	 * @returns The new or existing fellow instance
	 */
	static ensure(value: any, add?: boolean): Fellow
	/**
	 * Get a fellow from the singleton list
	 * @param value The value to fetch the value with
	 * @returns The fetched fellow, if they exist with that value
	 */
	static get(value: any): Fellow
	/**
	 * Add a fellow or a series of people, denoted by the value, to the singleton list
	 * @param values The fellow or people to add
	 * @returns An array of the fellow objects for the passed people
	 */
	static add(...values: any[]): Array<Fellow>
	/** Create a new Fellow instance with the value, however if the value is already a fellow instance, then just return it */
	static create(value: any): Fellow
	/**
	 * Construct our fellow instance with the value
	 * @param value The value used to set the properties of the fellow, forwarded to {@link Fellow::set}
	 */
	constructor(value: any)
	/**
	 * Update our fellow with the passed value
	 * @param fellow A string or object representation of the user
	 */
	set(fellow: string | Fellow | any): this
	/**
	 * If the name is empty, we will try to fallback to githubUsername then twitterUsername
	 * If the name is prefixed with a series of numbers, that is considered the year
	 * E.g. In `2015+ Bevry Pty Ltd` then `2015+` is the years
	 * E.g. In `2013-2015 Bevry Pty Ltd` then `2013-2015` is the years
	 */
	set name(value: string)
	/**
	 * Fetch the user's name, otherwise their usernames
	 */
	get name(): string
	/** Add the email to the set instead of replacing it */
	set email(value: string)
	/** Fetch the first email that was applied, otherwise an empty string */
	get email(): string
	/**
	 * Will determine if the passed URL is a github, facebook, or twitter URL.
	 * If it is, then it will extract the username and url from it.
	 * If it was not, then it will set the homepage variable.
	 */
	set url(input: string)
	/** Fetch the homepage with fallback to one of the service URLs if available */
	get url(): string
	/** Get the field field from the list that isn't empty */
	getFirstField(fields: string[]): any
	/** Get all fellows who administrate a particular repository */
	static administersRepository(repoSlug: string): Array<Fellow>
	/** Get all fellows who contribute to a particular repository */
	static contributesRepository(repoSlug: string): Array<Fellow>
	/** Get all fellows who maintain a particular repository */
	static maintainsRepository(repoSlug: string): Array<Fellow>
	/** Get all fellows who author a particular repository */
	static authorsRepository(repoSlug: string): Array<Fellow>
	/**
	 * Convert the fellow into the usual string format
	 * @example `NAME <EMAIL> (URL)`
	 */
	toString(format?: FormatOptions): string
	/**
	 * Convert the fellow into the usual markdown format
	 * @example `[NAME](URL) <EMAIL>`
	 */
	toMarkdown(format?: FormatOptions): string
	/**
	 * Convert the fellow into the usual HTML format
	 */
	toHTML(format?: FormatOptions): string
}
//# sourceMappingURL=index.d.ts.map
